#ifndef HANDEYE_H
#define HANDEYE_H

#include "opencv2/opencv.hpp"
#include <vector>

#if !defined RVCHANDEYE_EXPORT
#    if defined WIN32 && defined RVCHandEye_Detect_UsingShared
#        define RVCHANDEYE_API __declspec(dllimport)
#    else
#        define RVCHANDEYE_API
#    endif
#else
#    if defined WIN32
#        define RVCHANDEYE_API __declspec(dllexport)
#    else
#        define RVCHANDEYE_API __attribute__((visibility("default")))
#    endif
#endif

namespace RVCHandEye {

typedef enum {
    Err_OK = 0,
    Err_Algo_Marker_Fail_NoCircle = 101,
    Err_Algo_Marker_Fail_NoPoint = 102,
    Err_Algo_Marker_Fail_NoCirclePoint = 103,
    Err_Algo_Marker_Fail_ExceedOneCircle = 104,
    Err_Image_Empty = 105,
    Err_Invalid_Image2D = 106,
    Err_Invalid_ImagePoint = 107,
    Err_Notequal_Size = 108,

    Err_Algo_Solve_Fail = 201,
    Err_Param_Empty = 202,
    Err_Param_Size_Notequal = 203,
    Err_Param_Robot = 204,

    Err_Unkonwn = 1000

} ErrorCode;

/**
 * @brief 标定板坐标计算结果
 */
struct MarkerBoardResult {
    double x;
    double y;
    double z;

    double pixel_x;
    double pixel_y;
};

/**
 * @brief 误差列表
 * residual 总误差
 * residual x误差
 * residual y误差
 * residual z误差
 */
struct ResidualsList {
    double residual;
    double residual_x;
    double residual_y;
    double residual_z;
};

/**
 * @brief
 * max_t_error 最大误差
 * rms_t_error 均方根误差
 * rms_r_error 保留
 * max_r_error 保留
 * residuals 误差列表
 * valid_list 有效性列表, 当结果由暴力搜索得出时, 有效性列表与误差列表一一对应. 当valid_list[i]!=0时, residuals[i]的误差在最大误差之内,
 * 即为有效
 */
struct Residual {
    double rms_t_error;
    double max_t_error;
    double rms_r_error;
    double max_r_error;
    std::vector<ResidualsList> residuals;
    std::vector<int> valid_list;
};

/**
 * @brief EYE_IN_HAND:眼在手上, EYE_TO_HAND: 眼在手外
 */
typedef enum {
    EYE_IN_HAND,
    EYE_TO_HAND,
} HandEyeType;

/**
 * @brief 算法参数
 * type: 手眼类型
 * is_m: 坐标单位是否为m, is_m为true时, 单位为m, 否则为毫米
 * is_radian: 机器人旋转角度是否为弧度制, is_radian为true时，为弧度制，否则为角度值
 * exhaust: 是否暴力求解
 * ransac_dist: 最大误差
 */
struct AlgoParam {
    HandEyeType type;
    bool is_m;
    bool is_radian;
    bool exhaust;
    double ransac_dist;
};

/*
 * @brief 最终结果
 * x y z ox oz ow 保留
 * op 相机在基座下的位姿
 * matrix 矩阵
 * btms 标记在法兰下的位姿
 * quat 保留
 */
struct FinalResult {
    double x;
    double y;
    double z;
    double ox;
    double oz;
    double ow;

    std::vector<double> op;
    std::vector<std::vector<double>> matrix;
    Residual residual;
    std::vector<double> btms;
    std::vector<double> quat;
};
/**
 * @brief 获得SDK版本信息
 * @return
 */
RVCHANDEYE_API const char *GetVersion();

/**
 * @brief 解析错误代码的具体信息
 */
RVCHANDEYE_API const char *parseCode(const ErrorCode &code);

/**
 * @brief 计算标定板中心坐标
 * @param image2D 2D图
 * @param imagePoint 点云图
 * @param mbr 标定板坐标结果
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode SolveMarkerBoardResults(const cv::Mat &image2D, const cv::Mat &imagePoint, MarkerBoardResult &mbr);

/**
 * @brief 计算标定板中心坐标
 * @param image2DPath 2D图片文件路径
 * @param imagePointPath 3D点云文件路径
 * @param mbr 标定板坐标结果
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode SolveMarkerBoardResults(const std::string &image2DPath, const std::string &imagePointPath, MarkerBoardResult &mbr);

/**
 * @brief 计算最终标定结果, 注: 此求解函数同样适用于ABBRobot
 * @param markerResults 标定板坐标计算结果数组, 数组中每一项由`SolveMarkerBoardResults`计算得出
 * @param robotParams 机器人参数数组, 列表中每一项对应一个机器人参数，包含六项：X, Y, Z, OX, OY, OZ
 * @param algoParam 算法参数
 * @param fr 最终结果返回
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode KukaSolveFinalResult(const std::vector<MarkerBoardResult> &markerResults,
    const std::vector<std::vector<double>> &robotParams, const AlgoParam &algoParam, FinalResult &fr);

/**
 * @brief 计算最终标定结果, 注:此求解函数只适用于ABBQuatRobot
 * @param markerResults 标定板坐标计算结果数组, 数组中每一项由`SolveMarkerBoardResults`计算得出
 * @param robotParams 机器人参数数组, 列表中每一项对应一个机器人参数，包含七项：X, Y, Z, Q1, Q2, Q3, Q4
 * @param algoParam 算法参数
 * @param fr 最终结果返回
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode ABBQuatSolveFinalResult(const std::vector<MarkerBoardResult> &markerResults,
    const std::vector<std::vector<double>> &robotParams, const AlgoParam &algoParam, FinalResult &fr);

/**
 * @brief 计算最终标定结果, 注:此求解函数同样适用于YaskawaRobot
 * @param markerResults 标定板坐标计算结果数组, 数组中每一项由`SolveMarkerBoardResults`计算得出
 * @param robotParams 机器人参数数组, 列表中每一项对应一个机器人参数，包含六项：X, Y, Z, W, P, R
 * @param algoParam 算法参数
 * @param fr 最终结果返回
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode FanucSolveFinalResult(const std::vector<MarkerBoardResult> &markerResults,
    const std::vector<std::vector<double>> &robotParams, const AlgoParam &algoParam, FinalResult &fr);

/**
 * @brief 计算最终标定结果, 注:此求解函数只适用于UrRobot
 * @param markerResults 标定板坐标计算结果数组, 数组中每一项由`SolveMarkerBoardResults`计算得出
 * @param robotParams 机器人参数数组, 列表中每一项对应一个机器人参数，包含六项：X, Y, Z, RX, RY, RZ
 * @param algoParam 算法参数
 * @param fr 最终结果返回
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode UrSolveFinalResult(const std::vector<MarkerBoardResult> &markerResults,
    const std::vector<std::vector<double>> &robotParams, const AlgoParam &algoParam, FinalResult &fr);

/**
 * @brief 计算最终标定结果, 注:此求解函数只适用于通用类型Robot
 * @param markerResults 标定板坐标计算结果数组, 数组中每一项由`SolveMarkerBoardResults`计算得出
 * @param robotParams 机器人参数数组, 列表中每一项对应一个机器人参数，包含六项：X, Y, Z, R1, R2, R3
 * @param algoParam 算法参数
 * @param fr 最终结果返回
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode Common3SolveFinalResult(const std::vector<MarkerBoardResult> &markerResults,
    const std::vector<std::vector<double>> &robotParams, const AlgoParam &algoParam, FinalResult &fr);

/**
 * @brief 计算最终标定结果, 注:此求解函数只适用于通用类型Robot
 * @param markerResults 标定板坐标计算结果数组, 数组中每一项由`SolveMarkerBoardResults`计算得出
 * @param robotParams 机器人参数数组, 列表中每一项对应一个机器人参数，包含七项：X, Y, Z, R1, R2, R3, R4
 * @param algoParam 算法参数
 * @param fr 最终结果返回
 * @return 错误代码
 */
RVCHANDEYE_API ErrorCode Common4SolveFinalResult(const std::vector<MarkerBoardResult> &markerResults,
    const std::vector<std::vector<double>> &robotParams, const AlgoParam &algoParam, FinalResult &fr);

} // namespace RVCHandEye

#endif
