#include <pybind11/numpy.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <iostream>
#include <stdexcept>
#include <utility>

#include "HandEye.h"

using namespace RVCHandEye;
namespace py = pybind11;

#ifdef __GNUC__
#    define EXCEPTION(c) std::runtime_error(c)
#else
#    define EXCEPTION(c) std::exception(c)
#endif

PYBIND11_MODULE(PyHandEye, m) {
    py::enum_<ErrorCode>(m, "ErrorCode")
        .value("Err_OK", Err_OK)
        .value("Err_Algo_Marker_Fail_NoCircle", Err_Algo_Marker_Fail_NoCircle)
        .value("Err_Algo_Marker_Fail_NoPoint", Err_Algo_Marker_Fail_NoPoint)
        .value("Err_Algo_Marker_Fail_NoCirclePoint", Err_Algo_Marker_Fail_NoCirclePoint)
        .value("Err_Algo_Marker_Fail_ExceedOneCircle", Err_Algo_Marker_Fail_ExceedOneCircle)
        .value("Err_Image_Empty", Err_Image_Empty)
        .value("Err_Invalid_Image2D", Err_Invalid_Image2D)
        .value("Err_Invalid_ImagePoint", Err_Invalid_ImagePoint)
        .value("Err_Notequal_Size", Err_Notequal_Size)
        .value("Err_Algo_Solve_Fail", Err_Algo_Solve_Fail)
        .value("Err_Param_Empty", Err_Param_Empty)
        .value("Err_Param_Size_Notequal", Err_Param_Size_Notequal)
        .value("Err_Param_Robot", Err_Param_Robot)
        .value("Err_Unkonwn", Err_Unkonwn)
        .def(py::init([]() { return Err_OK; }));

    py::class_<MarkerBoardResult>(m, "MarkerBoardResult")
        .def(py::init([](const double x, const double y, const double z) {
            auto f = new MarkerBoardResult;
            f->x = x;
            f->y = y;
            f->z = z;
            return f;
        }),
            py::arg("x") = 0, py::arg("y") = 0, py::arg("z") = 0)
        .def_readwrite("x", &MarkerBoardResult::x)
        .def_readwrite("y", &MarkerBoardResult::y)
        .def_readwrite("z", &MarkerBoardResult::z)
        .def_readwrite("pixel_x", &MarkerBoardResult::pixel_x)
        .def_readwrite("pixel_y", &MarkerBoardResult::pixel_y);

    py::class_<ResidualsList>(m, "ResidualsList")
        .def(py::init())
        .def_readwrite("residual", &ResidualsList::residual)
        .def_readwrite("residual_x", &ResidualsList::residual_x)
        .def_readwrite("residual_y", &ResidualsList::residual_y)
        .def_readwrite("residual_z", &ResidualsList::residual_z);

    py::class_<Residual>(m, "Residual")
        .def(py::init())
        .def_readwrite("rms_t_error", &Residual::rms_t_error)
        .def_readwrite("max_t_error", &Residual::max_t_error)
        .def_readwrite("rms_r_error", &Residual::rms_r_error)
        .def_readwrite("max_r_error", &Residual::max_r_error)
        .def_readwrite("residuals", &Residual::residuals);

    py::enum_<HandEyeType>(m, "HandEyeType").value("EYE_IN_HAND", EYE_IN_HAND).value("EYE_TO_HAND", EYE_TO_HAND);

    py::class_<AlgoParam>(m, "AlgoParam")
        .def(py::init())
        .def_readwrite("type", &AlgoParam::type)
        .def_readwrite("is_m", &AlgoParam::is_m)
        .def_readwrite("is_radian", &AlgoParam::is_radian)
        .def_readwrite("exhaust", &AlgoParam::exhaust)
        .def_readwrite("ransac_dist", &AlgoParam::ransac_dist);

    py::class_<FinalResult>(m, "FinalResult")
        .def(py::init())
        .def_readwrite("x", &FinalResult::x)
        .def_readwrite("y", &FinalResult::y)
        .def_readwrite("z", &FinalResult::z)
        .def_readwrite("ox", &FinalResult::ox)
        .def_readwrite("oz", &FinalResult::oz)
        .def_readwrite("xw", &FinalResult::ow)
        .def_readwrite("op", &FinalResult::op)
        .def_readwrite("matrix", &FinalResult::matrix)
        .def_readwrite("residual", &FinalResult::residual)
        .def_readwrite("btms", &FinalResult::btms)
        .def_readwrite("quat", &FinalResult::quat);

    m.def("GetVersion", []() { return RVCHandEye::GetVersion(); });

    m.def("ParseCode", [](const ErrorCode &code) { return std::string(RVCHandEye::parseCode(code)); });

    m.def(
        "SolveMarkerBoardResults",
        [](const py::array_t<unsigned char> image2D, const py::array_t<double> imagePoint, const bool is_BGR) {
            cv::Mat i2D, iP;
            MarkerBoardResult mbr;
            if (image2D.ndim() == 3) {
                if (image2D.shape(2) == 3) {
                    cv::Mat tmp(image2D.shape(0), image2D.shape(1), CV_8UC3, const_cast<unsigned char *>(image2D.data()));
                    if (!is_BGR) {
                        cv::cvtColor(tmp, i2D, cv::COLOR_RGB2BGR);
                    } else {
                        tmp.copyTo(i2D);
                    }
                } else {
                    return std::pair<MarkerBoardResult, ErrorCode>(mbr, Err_Invalid_Image2D);
                }
            } else if (image2D.ndim() == 2) {
                cv::Mat tmp(image2D.shape(0), image2D.shape(1), CV_8UC1, const_cast<unsigned char *>(image2D.data()));
                tmp.copyTo(i2D);
            } else {
                return std::pair<MarkerBoardResult, ErrorCode>(mbr, Err_Invalid_Image2D);
            }

            iP = cv::Mat(image2D.shape(0), image2D.shape(1), CV_64FC3);
            if (imagePoint.ndim() > 3 || imagePoint.ndim() == 0) {
                return std::pair<MarkerBoardResult, ErrorCode>(mbr, Err_Invalid_ImagePoint);
            }
            if (imagePoint.size() / 3 != image2D.shape(0) * image2D.shape(1)) {
                return std::pair<MarkerBoardResult, ErrorCode>(mbr, Err_Notequal_Size);
            }

            memcpy(iP.data, imagePoint.data(), imagePoint.size() * imagePoint.itemsize());

            auto ret = SolveMarkerBoardResults(i2D, iP, mbr);
            if (ret != Err_OK) {
                return std::pair<MarkerBoardResult, ErrorCode>(mbr, ret);
            }
            return std::pair<MarkerBoardResult, ErrorCode>(mbr, Err_OK);
        },
        py::arg("image2D"), py::arg("imagePoint"), py::arg("is_BGR") = true);

    m.def(
        "SolveMarkerBoardResultsByPath",
        [](const std::string &img2DPath, const std::string &pointPath) {
            ErrorCode ret;
            MarkerBoardResult mbr;
            ret = SolveMarkerBoardResults(img2DPath, pointPath, mbr);
            return std::pair<MarkerBoardResult, ErrorCode>(mbr, ret);
        },
        py::arg("img2DPath"), py::arg("pointPath"));

    m.def(
        "KukaSolveFinalResult",
        [](const std::vector<MarkerBoardResult> &markerResults, const std::vector<std::vector<double>> &robotParams,
            const AlgoParam &algoParam) {
            FinalResult fr;
            auto ret = KukaSolveFinalResult(markerResults, robotParams, algoParam, fr);
            if (ret != Err_OK) {
                return std::pair<FinalResult, ErrorCode>(fr, ret);
            }
            return std::pair<FinalResult, ErrorCode>(fr, Err_OK);
        },
        py::arg("markerResults"), py::arg("robotParams"), py::arg("algoParam"));

    m.def(
        "ABBQuatSolveFinalResult",
        [](const std::vector<MarkerBoardResult> &markerResults, const std::vector<std::vector<double>> &robotParams,
            const AlgoParam &algoParam) {
            FinalResult fr;
            auto ret = ABBQuatSolveFinalResult(markerResults, robotParams, algoParam, fr);
            if (ret != Err_OK) {
                return std::pair<FinalResult, ErrorCode>(fr, ret);
            }
            return std::pair<FinalResult, ErrorCode>(fr, Err_OK);
        },
        py::arg("markerResults"), py::arg("robotParams"), py::arg("algoParam"));

    m.def(
        "FanucSolveFinalResult",
        [](const std::vector<MarkerBoardResult> &markerResults, const std::vector<std::vector<double>> &robotParams,
            const AlgoParam &algoParam) {
            FinalResult fr;
            auto ret = FanucSolveFinalResult(markerResults, robotParams, algoParam, fr);
            if (ret != Err_OK) {
                return std::pair<FinalResult, ErrorCode>(fr, ret);
            }
            return std::pair<FinalResult, ErrorCode>(fr, Err_OK);
        },
        py::arg("markerResults"), py::arg("robotParams"), py::arg("algoParam"));

    m.def(
        "UrSolveFinalResult",
        [](const std::vector<MarkerBoardResult> &markerResults, const std::vector<std::vector<double>> &robotParams,
            const AlgoParam &algoParam) {
            FinalResult fr;
            auto ret = UrSolveFinalResult(markerResults, robotParams, algoParam, fr);
            if (ret != Err_OK) {
                return std::pair<FinalResult, ErrorCode>(fr, ret);
            }
            return std::pair<FinalResult, ErrorCode>(fr, Err_OK);
        },
        py::arg("markerResults"), py::arg("robotParams"), py::arg("algoParam"));

    m.def(
        "Common3SolveFinalResult",
        [](const std::vector<MarkerBoardResult> &markerResults, const std::vector<std::vector<double>> &robotParams,
            const AlgoParam &algoParam) {
            FinalResult fr;
            auto ret = Common3SolveFinalResult(markerResults, robotParams, algoParam, fr);
            if (ret != Err_OK) {
                return std::pair<FinalResult, ErrorCode>(fr, ret);
            }
            return std::pair<FinalResult, ErrorCode>(fr, Err_OK);
        },
        py::arg("markerResults"), py::arg("robotParams"), py::arg("algoParam"));

    m.def(
        "Common4SolveFinalResult",
        [](const std::vector<MarkerBoardResult> &markerResults, const std::vector<std::vector<double>> &robotParams,
            const AlgoParam &algoParam) {
            FinalResult fr;
            auto ret = Common4SolveFinalResult(markerResults, robotParams, algoParam, fr);
            if (ret != Err_OK) {
                return std::pair<FinalResult, ErrorCode>(fr, ret);
            }
            return std::pair<FinalResult, ErrorCode>(fr, Err_OK);
        },
        py::arg("markerResults"), py::arg("robotParams"), py::arg("algoParam"));
}