import os
import re
import sys
import platform
import subprocess
import shutil
import site

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion
import warnings

def copy_files(source_dir, target_dir):
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    for file_name in os.listdir(source_dir):
        source_file = os.path.join(source_dir, file_name)
        target_file = os.path.join(target_dir, file_name)
        if os.path.isdir(source_file):
            copy_files(source_file, target_file)
        else:
            shutil.copy2(source_file, target_file)

if platform.system() == "Windows":
    filename = os.path.realpath(__file__)#C:\py\01.py
    dir = os.path.dirname(filename)#C:\py\
    if os.path.exists(dir + "/build"):
        shutil.rmtree(dir + "/build")
    copy_files(dir + '/../runtime', dir)

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        # cfg = 'Debug' if self.debug else 'Release'
        cfg = 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)
        if platform.system() == "Linux":
            try:
                subprocess.check_call(['patchelf', '--version'])
            except Exception as e:
                try:
                    subprocess.check_call(['apt', 'install', 'patchelf', '-y'])
                except subprocess.CalledProcessError as e:
                    warnings.warn(f"Can't install patchelf, check network please!")
                    return
            items = os.listdir(os.path.join(os.getcwd(), 'build'))
            targetDirs = [i for i in items if i.startswith('lib.')]
            if len(targetDirs) >=1:
                targetDir = os.path.join(os.getcwd(), 'build', targetDirs[0])
                files = os.listdir(targetDir)
                if len(files) ==1:
                    file = files[0]
                    targetFile = os.path.join(targetDir, file)
                    try:
                        subprocess.check_output(['patchelf', '--remove-rpath', f'{targetFile}'])
                        libAbsDir = os.path.abspath(os.path.join(os.getcwd(), '..', '..', 'lib'))
                        subprocess.check_output(['patchelf', '--force-rpath', '--set-rpath', f'{libAbsDir}', f'{targetFile}'])
                    except subprocess.CalledProcessError as e:
                        warnings.warn(f"Can't use patchelf or patchelf invalid! Output error is:{e.output}")
                        
                        
                    
            

setup(
    name='PyHandEye',
    version='2.0.0',
    author='rvbust',
    author_email='Unknown',
    package_data={
        "":["*.dll"]
    },
    packages=[""], 
    description='Rvbust HandEye Python SDK',
    long_description='',
    ext_modules=[CMakeExtension('PyHandEye')],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
)
